#ifndef MSP_GL_PROGRAMDATA_H_
#define MSP_GL_PROGRAMDATA_H_

#include <map>
#include <msp/datafile/objectloader.h>
#include "datatype.h"
#include "matrix.h"
#include "program.h"
#include "vector.h"

namespace Msp {
namespace GL {

class Buffer;
class Uniform;
class UniformBlock;
struct Color;

/**
Stores uniform variables for shader programs.  The uniforms are stored in a
program-independent way, and UniformBlocks are created to match the uniform
layouts of different programs.  If multiple programs have the same layout, the
same block is used for them.

The class is optimized for an access pattern where the set of uniforms and
programs stays constants, with only the values changing.
*/
class ProgramData
{
public:
	class Loader: public DataFile::ObjectLoader<ProgramData>
	{
	public:
		Loader(ProgramData &);
	private:
		void uniform1i(const std::string &, int);
		void uniform1f(const std::string &, float);
		void uniform2i(const std::string &, int, int);
		void uniform2f(const std::string &, float, float);
		void uniform3i(const std::string &, int, int, int);
		void uniform3f(const std::string &, float, float, float);
		void uniform4i(const std::string &, int, int, int, int);
		void uniform4f(const std::string &, float, float, float, float);
		void uniform_array_(const std::string &, DataType, unsigned);
		void uniform1i_array(const std::string &);
		void uniform1f_array(const std::string &);
		void uniform2i_array(const std::string &);
		void uniform2f_array(const std::string &);
		void uniform3i_array(const std::string &);
		void uniform3f_array(const std::string &);
		void uniform4i_array(const std::string &);
		void uniform4f_array(const std::string &);
		void uniform_array(const std::string &);
	};

private:
	class ArrayLoader: public DataFile::Loader
	{
	private:
		DataType type;
		unsigned element_size;
		std::vector<char> data;

	public:
		ArrayLoader(DataType, unsigned);

		DataType get_data_type() const { return type; }
		unsigned get_element_size() const { return element_size; }
		const void *get_data() const { return &data[0]; }
		unsigned get_size() const { return data.size()/(4*element_size); }

	private:
		void uniform(DataType, unsigned, const void *);
		void uniform1i(int);
		void uniform1f(float);
		void uniform2i(int, int);
		void uniform2f(float, float);
		void uniform3i(int, int, int);
		void uniform3f(float, float, float);
		void uniform4i(int, int, int, int);
		void uniform4f(float, float, float, float);
	};

	typedef unsigned Mask;

	enum
	{
		MASK_BITS = sizeof(Mask)*8,
		ALL_ONES = static_cast<Mask>(-1)
	};

	struct SharedBlock
	{
		Mask used;
		Mask dirty;
		UniformBlock *block;

		SharedBlock();
		SharedBlock(unsigned, UniformBlock *);
	};

	struct ProgramBlock
	{
		int bind_point;
		UniformBlock *block;
		SharedBlock *shared;

		ProgramBlock();
		ProgramBlock(int, SharedBlock *);
	};

	struct ProgramUniforms
	{
		std::vector<ProgramBlock> blocks;
		Mask used;
		Mask dirty;

		ProgramUniforms();
	};

	typedef std::map<std::string, unsigned> SlotMap;
	typedef std::map<Program::LayoutHash, SharedBlock> BlockMap;
	typedef std::map<Program::LayoutHash, ProgramUniforms> ProgramMap;

	// XXX All these mutables are a bit silly, but I'm out of better ideas
	const Program *tied_program;
	SlotMap uniform_slots;
	std::vector<Uniform *> uniforms;
	mutable BlockMap blocks;
	mutable ProgramMap programs;
	mutable UniformBlock *last_block;
	mutable Buffer *buffer;
	mutable unsigned dirty;

public:
	ProgramData(const Program * = 0);
	ProgramData(const ProgramData &);
	ProgramData &operator=(const ProgramData &);
	~ProgramData();

private:
	void uniform(const std::string &, Uniform *);
public:
	void uniform(const std::string &, const Uniform &);
	void uniform(const std::string &, int);
	void uniform(const std::string &, float);
	void uniform(const std::string &, int, int);
	void uniform(const std::string &, float, float);
	void uniform2(const std::string &, const int *);
	void uniform2(const std::string &, const float *);
	void uniform(const std::string &, int, int, int);
	void uniform(const std::string &, float, float, float);
	void uniform(const std::string &, const Vector3 &);
	void uniform3(const std::string &, const int *);
	void uniform3(const std::string &, const float *);
	void uniform(const std::string &, int, int, int, int);
	void uniform(const std::string &, float, float, float, float);
	void uniform(const std::string &, const Vector4 &);
	void uniform(const std::string &, const Color &);
	void uniform4(const std::string &, const int *);
	void uniform4(const std::string &, const float *);
	void uniform(const std::string &, const LinAl::Matrix<float, 2, 2> &);
	void uniform_matrix2(const std::string &, const float *);
	void uniform(const std::string &, const LinAl::Matrix<float, 3, 3> &);
	void uniform_matrix3(const std::string &, const float *);
	void uniform(const std::string &, const Matrix &);
	void uniform_matrix4(const std::string &, const float *);
	void uniform1_array(const std::string &, unsigned, const int *);
	void uniform1_array(const std::string &, unsigned, const float *);
	void uniform2_array(const std::string &, unsigned, const int *);
	void uniform2_array(const std::string &, unsigned, const float *);
	void uniform3_array(const std::string &, unsigned, const int *);
	void uniform3_array(const std::string &, unsigned, const float *);
	void uniform4_array(const std::string &, unsigned, const int *);
	void uniform4_array(const std::string &, unsigned, const float *);
	void uniform_matrix4_array(const std::string &, unsigned, const float *);
	void remove_uniform(const std::string &);

private:
	unsigned compute_slot_mask(const Program::UniformBlockInfo &) const;
	void update_block(UniformBlock &, const Program::UniformBlockInfo &) const;
	SharedBlock *get_shared_block(const Program::UniformBlockInfo &) const;

public:
	/** Applies uniform blocks for the currently bound program, creating them
	if needed. */
	void apply() const;
};

} // namespace GL
} // namespace Msp

#endif
